@@ -2,14 +2,16 @@
 from __future__ import unicode_literals
 
 from collections import defaultdict
+import socket
 
 # Import third party libs
 from librouteros import connect
 from librouteros.exceptions import TrapError
 from librouteros.exceptions import FatalError
-from librouteros.exceptions import ConnectionError
 from librouteros.exceptions import MultiTrapError
 import librouteros.login
+from librouteros.query import Key
+from netaddr import IPNetwork
 
 # Import NAPALM base
 from napalm.base import NetworkDriver
@@ -24,28 +26,26 @@ from napalm_ros.utils import to_seconds
 from napalm_ros.utils import iface_addresses
 
 
+# pylint: disable=too-few-public-methods,undefined-variable
+class Keys:
+    bgp = Key('bgp')
+    dst_addr = Key('dst-address')
+    rcv_from = Key('received-from')
+
+
+# pylint: disable=too-many-public-methods
 class ROSDriver(NetworkDriver):
 
+    # pylint: disable=super-init-not-called
     def __init__(self, hostname, username, password, timeout=60, optional_args=None):
         self.hostname = hostname
         self.username = username
         self.password = password
         self.timeout = timeout
-        optional_args = optional_args or dict()
-        self._process_optional_args(optional_args)
-        self.port = optional_args.get('port', 8728)
+        self.optional_args = optional_args or dict()
+        self.port = self.optional_args.get('port', 8728)
         self.api = None
 
-    def _process_optional_args(self, optional_args={}):
-        self.login_methods = tuple(
-            set(
-                [
-                    getattr(librouteros.login, method, "login_plain")
-                    for method in optional_args.get("login_methods", ["login_token", "login_plain"])
-                ]
-            )
-        )
-
     def close(self):
         self.api.close()
 
@@ -58,10 +58,10 @@ class ROSDriver(NetworkDriver):
         for iface in self.api('/interface/print', stats=True):
             result[iface['name']] = defaultdict(int)
             stats = result[iface['name']]
-            stats['tx_errors'] += iface['tx-error']
-            stats['rx_errors'] += iface['rx-error']
-            stats['tx_discards'] += iface['tx-drop']
-            stats['rx_discards'] += iface['rx-drop']
+            stats['tx_errors'] += iface.get('tx-error', 0)
+            stats['rx_errors'] += iface.get('rx-error', 0)
+            stats['tx_discards'] += iface.get('tx-drop', 0)
+            stats['rx_discards'] += iface.get('rx-drop', 0)
             stats['tx_octets'] += iface['tx-byte']
             stats['rx_octets'] += iface['rx-byte']
             stats['tx_unicast_packets'] += iface['tx-packet']
@@ -74,11 +74,119 @@ class ROSDriver(NetworkDriver):
 
         return result
 
+    # pylint: disable=invalid-name
+    def get_bgp_neighbors(self):
+        bgp_neighbors = defaultdict(lambda: dict(peers={}))
+        sent_prefixes = defaultdict(lambda: defaultdict(int))
+
+        # Count prefixes advertised to each configured peer
+        for route in self.api("/routing/bgp/advertisements/print"):
+            sent_prefixes[route["peer"]]["ipv{}".format(IPNetwork(route["prefix"]).version)] += 1
+        # Calculate stats for each routing bgp instance
+        for inst in self.api("/routing/bgp/instance/print"):
+            instance_name = "global" if inst["name"] == "default" else inst["name"]
+            bgp_neighbors[instance_name]["router_id"] = inst["router-id"]
+            inst_peers = find(self.api("/routing/bgp/peer/print"), key="instance", value=inst["name"])
+            for peer in inst_peers:
+                prefix_stats = {}
+                # Mikrotik prefix counts are not per-AFI so attempt to query
+                # the routing table if more than one address family is present on a peer
+                if len(peer["address-families"].split(",")) > 1:
+                    for af in peer["address-families"].split(","):
+                        prefix_count = len(self.api.path(f"/{af}/route").select(dst_addr).where(
+                            Keys.bgp == True, # pylint: disable=singleton-comparison
+                            Keys.rcv_from == peer["name"],
+                        ))
+                        family = "ipv4" if af == "ip" else af
+                        prefix_stats[family] = {
+                            "sent_prefixes": sent_prefixes.get(peer["name"], {}).get(family, 0),
+                            "accepted_prefixes": prefix_count,
+                            "received_prefixes": prefix_count,
+                        }
+                else:
+                    family = "ipv4" if peer["address-families"] == "ip" else af
+                    prefix_stats[family] = {
+                        "sent_prefixes": sent_prefixes.get(peer["name"], {}).get(family, 0),
+                        "accepted_prefixes": peer.get("prefix-count", 0),
+                        "received_prefixes": peer.get("prefix-count", 0),
+                    }
+                bgp_neighbors[instance_name]["peers"][peer["remote-address"]] = {
+                    "local_as": inst["as"],
+                    "remote_as": peer["remote-as"],
+                    "remote_id": peer.get("remote-id", ""),
+                    "is_up": peer.get("established", False),
+                    "is_enabled": not peer["disabled"],
+                    "description": peer["name"],
+                    "uptime": to_seconds(peer.get("uptime", "0s")),
+                    "address_family": prefix_stats,
+                }
+        return dict(bgp_neighbors)
+
+    def get_bgp_neighbors_detail(self, neighbor_address=""):
+        sent_prefixes = defaultdict(int)
+
+        bgp_neighbors = defaultdict(lambda: defaultdict(list))
+        # Get all configured peers
+        peers = self.api("/routing/bgp/peer/print")
+        # Get all configured advertisements
+        advertisements = self.api("/routing/bgp/advertisements/print")
+        # Count prefixes advertised to each configured peer
+        for route in advertisements:
+            sent_prefixes[route["peer"]] += 1
+
+        for inst in self.api("/routing/bgp/instance/print"):
+            instance_name = "global" if inst["name"] == "default" else inst["name"]
+            inst_peers = find(peers, key="instance", value=inst["name"])
+
+            for peer in inst_peers:
+                if neighbor_address and peer["remote-address"] != neighbor_address:
+                    continue
+                peer_details = {
+                    "up": peer.get("established", False),
+                    "local_as": inst["as"],
+                    "remote_as": peer["remote-as"],
+                    "router_id": inst["router-id"],
+                    "local_address": peer.get("local-address", False),
+                    "local_address_configured": bool(peer.get("local-address", False)),
+                    "local_port": 179,
+                    "routing_table": inst["routing-table"],
+                    "remote_address": peer["remote-address"],
+                    "remote_port": 179,
+                    "multihop": peer["multihop"],
+                    "multipath": False,
+                    "remove_private_as": peer["remove-private-as"],
+                    "import_policy": peer["in-filter"],
+                    "export_policy": peer["out-filter"],
+                    "input_messages": peer.get("updates-received", 0) + peer.get("withdrawn-received", 0),
+                    "output_messages": peer.get("updates-sent", 0) + peer.get("withdrawn-sent", 0),
+                    "input_updates": peer.get("updates-received", 0),
+                    "output_updates": peer.get("updates-sent", 0),
+                    "messages_queued_out": 0,
+                    "connection_state": peer.get("state", ""),
+                    "previous_connection_state": "",
+                    "last_event": "",
+                    "suppress_4byte_as": not peer.get("as4-capability", True),
+                    "local_as_prepend": False,
+                    "holdtime": to_seconds(peer.get("used-hold-time", peer.get("hold-time", "30s"))),
+                    "configured_holdtime": to_seconds(peer.get("hold-time", "30s")),
+                    "keepalive": to_seconds(peer.get("used-keepalive-time", "10s")),
+                    "configured_keepalive": to_seconds(peer.get("keepalive-time", "10s")),
+                    "active_prefix_count": peer.get("prefix-count", 0),
+                    "received_prefix_count": peer.get("prefix-count", 0),
+                    "accepted_prefix_count": peer.get("prefix-count", 0),
+                    "suppressed_prefix_count": 0,
+                    "advertised_prefix_count": sent_prefixes.get(peer["name"], 0),
+                    "flap_count": 0,
+                }
+                bgp_neighbors[instance_name][peer["remote-as"]].append(peer_details)
+
+        return {k: dict(v) for k, v in bgp_neighbors.items()}
+
     def get_arp_table(self, vrf=""):
         if vrf:
             vrfs = self.api('/ip/route/vrf/print')
             vrfs = find(vrfs, key='routing-mark', value=vrf)
-            interfaces = tuple(splitKey(vrfs, 'interfaces'))
+            interfaces = tuple(split_key(vrfs, 'interfaces'))
             arp_table = list(entry for entry in self.arp if entry['interface'] in interfaces)
         else:
             arp_table = list(self.arp)
@@ -87,18 +195,37 @@ class ROSDriver(NetworkDriver):
 
     def get_mac_address_table(self):
         table = list()
-        for entry in self.api('/interface/ethernet/switch/unicast-fdb/print'):
+        for entry in self.api('/interface/bridge/host/print'):
             table.append(
                 dict(
                     mac=entry['mac-address'],
-                    interface=entry['port'],
-                    vlan=entry['vlan-id'],
+                    interface=entry['interface'],
+                    vlan=entry.get('vid', 1),     # The vid is not consistently set in the API
                     static=not entry['dynamic'],
-                    active=entry['active'],
+                    active=not entry['invalid'],
                     moves=0,
                     last_move=0.0,
                 )
             )
+
+        try:
+            for entry in self.api('/interface/ethernet/switch/unicast-fdb/print'):
+                table.append(
+                    dict(
+                        mac=entry['mac-address'],
+                        interface=entry['port'],
+                        vlan=entry['vlan-id'],
+                        static=not entry['dynamic'],
+                        active=entry['active'],
+                        moves=0,
+                        last_move=0.0,
+                    )
+                )
+        except librouteros.exceptions.TrapError:
+            # This only exists in the CRS1XX and CRS2XX switches.
+            # Ignore if not present on the current device.
+            pass
+
         return table
 
     def get_network_instances(self, name=""):
@@ -159,13 +286,13 @@ class ROSDriver(NetworkDriver):
         for entry in self.api('/ip/arp/print'):
             if 'mac-address' not in entry:
                 continue
-            else:
-                yield {
-                    'interface': entry['interface'],
-                    'mac': cast_mac(entry['mac-address']),
-                    'ip': cast_ip(entry['address']),
-                    'age': float(-1),
-                }
+
+            yield {
+                'interface': entry['interface'],
+                'mac': cast_mac(entry['mac-address']),
+                'ip': cast_ip(entry['address']),
+                'age': float(-1),
+            }
 
     def get_ipv6_neighbors_table(self):
         ipv6_neighbors_table = []
@@ -196,7 +323,7 @@ class ROSDriver(NetworkDriver):
         }
 
         try:
-            system_health = self.api('/system/health/print')[0]
+            system_health = tuple(self.api('/system/health/print'))[0]
         except IndexError:
             return environment
 
@@ -225,7 +352,7 @@ class ROSDriver(NetworkDriver):
             }
 
         try:
-            system_resource = self.api('/system/resource/print')[0]
+            system_resource = tuple(self.api('/system/resource/print'))[0]
         except IndexError:
             return dict()
 
@@ -239,10 +366,10 @@ class ROSDriver(NetworkDriver):
         return environment
 
     def get_facts(self):
-        resource = self.api('/system/resource/print')[0]
-        identity = self.api('/system/identity/print')[0]
-        routerboard = self.api('/system/routerboard/print')[0]
-        interfaces = self.api('/interface/print')
+        resource = tuple(self.api('/system/resource/print'))[0]
+        identity = tuple(self.api('/system/identity/print'))[0]
+        routerboard = tuple(self.api('/system/routerboard/print'))[0]
+        interfaces = tuple(self.api('/interface/print'))
         return {
             'uptime': to_seconds(resource['uptime']),
             'vendor': resource['platform'],
@@ -251,7 +378,9 @@ class ROSDriver(NetworkDriver):
             'fqdn': u'',
             'os_version': resource['version'],
             'serial_number': routerboard.get('serial-number', ''),
-            'interface_list': napalm.base.utils.string_parsers.sorted_nicely(tuple(iface['name'] for iface in interfaces)),
+            'interface_list': napalm.base.utils.string_parsers.sorted_nicely(
+                tuple(iface['name'] for iface in interfaces),
+            ),
         }
 
     def get_interfaces(self):
@@ -271,13 +400,13 @@ class ROSDriver(NetworkDriver):
     def get_interfaces_ip(self):
         interfaces_ip = {}
 
-        ipv4_addresses = self.api('/ip/address/print')
+        ipv4_addresses = tuple(self.api('/ip/address/print'))
         for ifname in (row['interface'] for row in ipv4_addresses):
             interfaces_ip.setdefault(ifname, dict())
             interfaces_ip[ifname]['ipv4'] = iface_addresses(ipv4_addresses, ifname)
 
         try:
-            ipv6_addresses = self.api('/ipv6/address/print')
+            ipv6_addresses = tuple(self.api('/ipv6/address/print'))
             for ifname in (row['interface'] for row in ipv6_addresses):
                 interfaces_ip.setdefault(ifname, dict())
                 interfaces_ip[ifname]['ipv6'] = iface_addresses(ipv6_addresses, ifname)
@@ -288,7 +417,7 @@ class ROSDriver(NetworkDriver):
 
     def get_ntp_servers(self):
         ntp_servers = {}
-        ntp_client_values = self.api('/system/ntp/client/print')[0]
+        ntp_client_values = tuple(self.api('/system/ntp/client/print'))[0]
         fqdn_ntp_servers = filter(None, ntp_client_values.get('server-dns-names', '').split(','))
         for ntp_peer in fqdn_ntp_servers:
             ntp_servers[ntp_peer] = {}
@@ -308,7 +437,7 @@ class ROSDriver(NetworkDriver):
                 'mode': u'ro' if row.get('read-access') else 'rw',
             }
 
-        snmp_values = self.api('/snmp/print')[0]
+        snmp_values = tuple(self.api('/snmp/print'))[0]
 
         return {
             'chassis_id': snmp_values['engine-id'],
@@ -324,17 +453,22 @@ class ROSDriver(NetworkDriver):
         return users
 
     def open(self):
+        method = self.optional_args.get('login_method', 'plain')
+        method = getattr(librouteros.login, method)
         try:
             self.api = connect(
                 host=self.hostname,
                 username=self.username,
                 password=self.password,
+                port=self.port,
                 timeout=self.timeout,
-                login_methods=self.login_methods,
+                login_method=method,
             )
-        except (TrapError, FatalError, ConnectionError, MultiTrapError) as exc:
+        except (TrapError, FatalError, socket.timeout, socket.error, MultiTrapError) as exc:
+            # pylint: disable=raise-missing-from
             raise ConnectionException("Could not connect to {}:{} - [{!r}]".format(self.hostname, self.port, exc))
 
+    # pylint: disable=too-many-arguments
     def ping(
         self,
         destination,
@@ -346,7 +480,6 @@ class ROSDriver(NetworkDriver):
         vrf=C.PING_VRF
     ):
         params = {
-            'count': count,
             'address': destination,
             'ttl': ttl,
             'size': size,
@@ -357,13 +490,13 @@ class ROSDriver(NetworkDriver):
         if vrf:
             params['routing-table'] = vrf
 
-        results = self.api('/ping', **params)
-
+        results = tuple(self.api('/ping', **params))
+        rtt = lambda x: (float(row.get(x, '-1ms').replace('ms', '')) for row in results)
         ping_results = {
             'probes_sent': max(row['sent'] for row in results),
             'packet_loss': max(row['packet-loss'] for row in results),
-            'rtt_min': min(float(row.get('min-rtt', '-1ms').replace('ms', '')) for row in results),
-            'rtt_max': max(float(row.get('max-rtt', '-1ms').replace('ms', '')) for row in results), # Last result has calculated avg
+            'rtt_min': min(rtt('min-rtt')),
+            'rtt_max': max(rtt('max-rtt')),                                         # Last result has calculated avg
             'rtt_avg': float(results[-1].get('avg-rtt', '-1ms').replace('ms', '')),
             'rtt_stddev': float(-1),
             'results': []
@@ -377,10 +510,6 @@ class ROSDriver(NetworkDriver):
 
         return dict(success=ping_results)
 
-    def _system_package_enabled(self, package):
-        enabled = (pkg['name'] for pkg in self.api('/system/package/print') if not pkg['disabled'])
-        return package in enabled
-
 
 def find(haystack, key, value):
     for row in haystack:
@@ -388,7 +517,7 @@ def find(haystack, key, value):
             yield row
 
 
-def splitKey(haystack, key):
+def split_key(haystack, key):
     for row in haystack:
         for item in row[key].split(','):
             yield item
